from dataclasses import dataclass
import json
import re
from typing import List
from lxml import html
from lxml.etree import ElementBase
from datetime import datetime, timedelta

from .interfaces.AioHttpHelper import AioHttpHelperInterface
from .kwargs import extract_common_kwargs
from .util import typedxpath


@dataclass
class CodeforcesUser:
  name: str
  title: str
  class__: str
  profile: str


@dataclass
class ContestListItem:
  id: int
  title: str
  authors: List[CodeforcesUser]
  start: int  # timestamp
  length: str
  participants: str
  upcoming: bool
  # only for upcoming
  registered: bool
  # ['D','1','2'], ['C']
  Div: List[str]


def parse_div(title: str) -> List[str]:
  r: List[str] = []
  if 'Div.' in title:
    r += ['D']
    for i in range(1, 5):
      if re.compile('Div\\. ?' + str(i)).search(title): r += [str(i)]
  elif not 'unrated' in title.lower():
    r += ['C']
  return r


# 获取已经解决的 统计
async def async_solved_count(http: AioHttpHelperInterface, solved_path: str):
  solved_string = await http.async_post("/data/contests", {'action': 'getSolvedProblemCountsByContest'}, csrf=True)
  solved_json = json.loads(solved_string)
  open(solved_path, 'w').write(solved_string)
  return solved_json


# [day:]hour:minutes
def ddhhmm2seconds(length: str) -> int:
  s = length.split(':')
  days = int(s[0]) if len(s) == 3 else 0
  hours = int(s[-2])
  minutes = int(s[-1])
  return int(timedelta(days=days, hours=hours, minutes=minutes).total_seconds())


def is_contest_running(item: ContestListItem) -> bool:
  start = item.start
  end = start + ddhhmm2seconds(item.length)
  now = int(datetime.now().timestamp())  # local time zone
  return now >= start and now < end


@dataclass
class ContestList:
  upcomming: List[ContestListItem]
  history: List[ContestListItem]


def parse_contest_list(raw_contests: ElementBase, upcoming: bool, **kw) -> List[ContestListItem]:
  logger = extract_common_kwargs(**kw).logger
  contests: List[ContestListItem] = []

  for c in typedxpath(raw_contests, './/tr[@data-contestid]'):
    try:
      cid = int(c.get('data-contestid'))
      td = typedxpath(c, './/td')
      title = td[0].text.lstrip().splitlines()[0]
      authors: List[CodeforcesUser] = [
          CodeforcesUser(
              class__=a.get('class').split(' ')[1],
              profile=a.get('href'),
              title=a.get('title'),
              name=a.text,
          ) for a in typedxpath(td[1], './/a')
      ]
      start_str = typedxpath(td[2], './/span')[0].text
      start = int(datetime.strptime(str(start_str) + "+0300", "%b/%d/%Y %H:%M%z").timestamp())  # Russian + 3hours
      length = td[3].text.strip()
      participants = ''
      registration = False

      if upcoming:
        msg = td[5].text_content().strip()
        if msg.startswith("Registration completed"):
          registration = True
        participants = re.sub('\\s+', ' ', msg.split('x')[-1])
      else:
        participants = re.sub('\\s+', ' ', td[5].text_content().strip().lstrip('x'))

      contests.append(
          ContestListItem(id=cid,
                          title=title,
                          authors=authors,
                          start=start,
                          length=length,
                          participants=participants,
                          registered=registration,
                          upcoming=upcoming,
                          Div=parse_div(title)))
    except Exception as e:
      logger.exception(e)
  return contests


# This function is to simulate web request, do not do the cache
async def async_contest_list(http: AioHttpHelperInterface, page: int = 1, **kw) -> ContestList:
  """
    This method will use ``http`` for get contests page, you can both login or not login

    :param page: the page in url

    :returns: the result

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.contest_list import async_contest_list

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          # you can login before get list
          result = await async_contest_list(http=http)
          for c in result.upcomming[:5]:
              print(c)
          for c in result.history[:5]:
              print(c)
          await http.close_session()

        asyncio.run(demo())
  """
  logger = extract_common_kwargs(**kw).logger
  doc = html.fromstring(await http.async_get(f'/contests/page/{page}?complete=true'))
  table = typedxpath(doc, './/div[@class="datatable"]')
  upcoming = parse_contest_list(table[0], upcoming=True)

  # count contests
  if len(table[1].xpath('.//tr[@data-contestid]')) == 0:
    logger.error("[!] Contest is running or countdown")
  else:
    history = parse_contest_list(table[1], upcoming=False)

  if not history:
    logger.error("? list1 is empty")
    return ContestList(upcomming=upcoming, history=[])
  return ContestList(upcomming=upcoming, history=history)



from dataclasses import dataclass
from typing import List
from bs4 import BeautifulSoup
import bs4

from .interfaces.AioHttpHelper import AioHttpHelperInterface


@dataclass
class Lang:
  text: str
  value: str


async def async_language(http: AioHttpHelperInterface, **kw) -> List[Lang]:
  """
    This method will use ``http`` to request ``/problemset/submit``, and parse language options

    :returns: parsed structured result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.account import async_login
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.language import async_language

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          result = await async_login(http=http, handle='<handle>', password='<password>')
          assert(result.success)
          result = await async_language(http)
          for item in result:
            print(item)
          await http.close_session()

        asyncio.run(demo())
  """
  # login before
  resp = await http.async_get(f'/problemset/submit')
  ret: List[Lang] = []
  soup = BeautifulSoup(resp, 'lxml')
  tags = soup.find('select', attrs={'name': 'programTypeId'})
  if isinstance(tags, bs4.Tag):
    for child in tags.find_all('option'):
      ret.append(Lang(value=child.get('value'), text=child.string))
  return ret



from collections import defaultdict
from dataclasses import dataclass, field
from os import path
from typing import Any, List, Tuple
from lxml import html

# from .ui import BLUE, GREEN, RED, redraw
from .util import typedxpath
from .account import is_user_logged_in
from .interfaces.AioHttpHelper import AioHttpHelperInterface
from .kwargs import extract_common_kwargs
from .url import problem_url_parse


async def async_submit(http: AioHttpHelperInterface, contest_id: str, level: str, file_path: str, lang_id: str,
                       **kw) -> Tuple[str, str]:
  """
    This method will use ``http`` to post submit

    :param http: AioHttpHelperInterface 
    :param ws_handler: function to handler messages 

    :returns: (submission_id, html_text of contest/<contest id>/my )

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.account import async_login
        from codeforces_core.websocket import create_contest_ws_task
        from codeforces_core.submit import async_submit, display_contest_ws

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          result = await async_login(http=http, handle='<handle>', password='<password>')
          assert(result.success)

          print('before submit')
          submit_id, resp = await async_submit(http, contest_id='1777', level='F', file_path='F.cpp', lang_id='73')
          print('submit id:',submit_id)

          # connect websocket before submit sometimes cannot receive message
          contest_task = create_contest_ws_task(http, contest_id='1777', ws_handler=display_contest_ws)
          print("contest ws created");

          try:
            result = await asyncio.wait_for(contest_task, timeout=30)
            print("ws is done, result:", result)
          except asyncio.TimeoutError:
            pass
          await http.close_session()

        asyncio.run(demo())

  """
  logger = extract_common_kwargs(**kw).logger

  if not contest_id or not level:
    logger.error("[!] Invalid contestID or level")
    return '', ''
  if not path.isfile(file_path):
    logger.error("[!] File not found : {}".format(file_path))
    return '', ''

  token = http.get_tokens()
  submit_form = {
      'csrf_token': token['csrf'],
      'ftaa': token['ftaa'],
      'bfaa': token['bfaa'],
      'action': 'submitSolutionFormSubmitted',
      'submittedProblemIndex': level,
      'programTypeId': lang_id,
  }
  url = '/contest/{}/problem/{}?csrf_token={}'.format(contest_id, level.upper(), token['csrf'])
  form = http.create_form(submit_form)
  form.add_field('sourceFile', open(file_path, 'rb'), filename=file_path)
  resp = await http.async_post(url, form)  # 正常是 302 -> https://codeforces.com/contest/<contest id>/my
  if not is_user_logged_in(resp):
    logger.error("Login required")
    return '', resp
  doc = html.fromstring(resp)
  for e in typedxpath(doc, './/span[@class="error for__sourceFile"]'):
    if e.text == 'You have submitted exactly the same code before':
      logger.error("[!] " + e.text)
      return '', resp

  status = parse_submit_status(resp)[0]
  assert status.url.split('/')[-1] == level.upper()
  return status.id, resp


# TODO move oiterminal code to here use dataclass
@dataclass
class SubmissionPageResult:
  id: str = ''
  url: str = ''
  verdict: str = ''
  time_ms: str = ''
  mem_bytes: str = ''


# status_url = f'/contest/{contest_id}/my'
# resp = await http.async_get(status_url)
# status = parse_submit_status(resp)
def parse_submit_status(html_page: str) -> List[SubmissionPageResult]:
  ret: List[SubmissionPageResult] = []
  doc = html.fromstring(html_page)
  tr = typedxpath(doc, './/table[@class="status-frame-datatable"]/tr[@data-submission-id]')
  for t in tr:
    td = t.xpath('.//td')
    submission_id = ''.join(td[0].itertext()).strip()
    url = td[3].xpath('.//a[@href]')[0].get('href')
    verdict = ''.join(td[5].itertext()).strip()
    prog_time = td[6].text.strip().replace('\xa0', ' ').split()[0]
    prog_mem = td[7].text.strip().replace('\xa0', ' ').split()[0]
    ret.append(SubmissionPageResult(id=submission_id, url=url, verdict=verdict, time_ms=prog_time, mem_bytes=prog_mem))
  return ret


async def async_fetch_submission_page(http: AioHttpHelperInterface, problem_url: str,
                                      **kw) -> List[SubmissionPageResult]:
  contest_id, problem_key = problem_url_parse(problem_url)
  # 正常是 302 -> https://codeforces.com/contest/<contest id>/my
  html_page = await http.async_get(f'/contest/{contest_id}/my')
  result = parse_submit_status(html_page)
  return list(filter(lambda o: o.url.endswith(problem_key), result))


@dataclass
class SubmissionWSResult:
  source: Any = field(default_factory=lambda: defaultdict(dict))
  submit_id: int = 0
  contest_id: int = 0
  title: str = ''
  msg: str = ''
  passed: int = 0
  testcases: int = 0
  ms: int = 0
  mem: int = 0
  date1: str = ''
  date2: str = ''
  lang_id: int = 0


# TODO 两个不同的ws(公共的和针对题目的) 似乎返回结构不同
def transform_submission(data: Any) -> SubmissionWSResult:
  d = data['text']['d']
  return SubmissionWSResult(
      source=data,
      # [5973095143352889425, ???? data-a
      submit_id=d[1],  # 200625609,
      contest_id=d[2],  # 1777,
      # 1746206, ??
      title=d[4],  # 'TESTS',
      # None,
      msg=d[6],  # 'TESTING', 'OK'
      passed=d[7],  # 0, ??
      testcases=d[8],  # 81, ?? 在测试过程中 这个值会增长,而d[7]一直是0,直到'OK'
      ms=d[9],  # 0,
      mem=d[10],  # 0, Bytes
      # 148217099,
      # '215020',
      date1=d[13],  # '04.04.2023 3:21:48',
      date2=d[14],  # '04.04.2023 3:21:48',
      # 2147483647,
      lang_id=d[16],  # 73,
      # 0]
  )


# return (end watch?, transform result)
def display_contest_ws(result: Any) -> Tuple[bool, Any]:
  parsed_data = transform_submission(result)
  print(parsed_data)
  if parsed_data.msg != 'TESTING':
    return True, parsed_data
  return False, parsed_data



from dataclasses import dataclass, field
from enum import Enum
from typing import List, cast
from bs4 import BeautifulSoup
import bs4
from lxml import html
from lxml.etree import _Element
from lxml.html import HtmlMixin

from .interfaces.AioHttpHelper import AioHttpHelperInterface
from .util import soup_find_bs4Tag


class E_STATUS(str, Enum):
  NOT_SUBMITTED = ""
  AC = "AC"
  ERROR = "ERROR"


@dataclass
class ProblemMeta:
  id: str = ''
  url: str = ''
  name: str = ''
  passed: str = ''  # number of passed submission in contest
  status: E_STATUS = E_STATUS.NOT_SUBMITTED
  time_limit_msec: int = 0
  memory_limit_kb: int = 0
  contest_id: str = ''


@dataclass
class ContestMeta:
  id: str = ''
  url: str = ''
  problems: List[ProblemMeta] = field(default_factory=lambda: [])


@dataclass
class Questions:
  when: str
  Q: str
  A: str


@dataclass
class Materials:
  text: str
  url: str


@dataclass
class ContestPage:
  id: str
  url: str
  title: str
  problems: List[ProblemMeta]
  # questions: List[Questions]
  materials: List[Materials]


def parse_problems(resp: str) -> List[ProblemMeta]:
  soup = BeautifulSoup(resp, 'lxml')
  currentContestList = soup.find('div', class_='datatable')

  assert isinstance(currentContestList, bs4.Tag)
  trs: List[BeautifulSoup] = currentContestList.find_all('tr')
  ret = []
  for i in range(1, len(trs)):
    tds = trs[i].find_all('td')
    namediv = tds[1].div.div
    row = ProblemMeta(
        id=tds[0].get_text().strip(),
        url=tds[0].find('a')['href'],
        name=namediv.get_text().strip(),
        passed=tds[3].get_text().split('x')[-1].strip(),
    )

    if 'class' in trs[i].attrs:
      classes: str = trs[i].attrs['class']  # mypy ?, List[str]
      if 'accepted-problem' in classes:
        row.status = E_STATUS.AC
      elif 'rejected-problem' in classes:
        row.status = E_STATUS.ERROR
    limit = tds[1].find('div', class_="notice").get_text().strip()
    if limit.startswith('standard input/output'):
      limit = limit[len('standard input/output'):].strip()
      try:
        t, m = limit.split(',')
        row.time_limit_msec = int(float(t.strip().split(' ')[0]) * 1000)
        row.memory_limit_kb = int(float(m.strip().split(' ')[0]) * 1000)
      except:
        pass

    ret.append(row)
  return ret


def parse_materials(resp: str) -> List[Materials]:
  doc = html.fromstring(resp)
  captions = cast(List[_Element], doc.xpath('.//div[@class="caption titled"]'))
  ret = []
  for c in captions:
    title = c.text[1:].strip()
    if title != 'Contest materials':
      continue
    links = cast(List[_Element], c.getparent().xpath('.//a[@href]'))
    for a in links:
      title_text = str(cast(HtmlMixin, html.fromstring(a.get('title'))).text_content())
      ret.append(Materials(text=title_text, url=a.get('href')))
  return ret


async def async_contest_meta(http: AioHttpHelperInterface, contest_id: str, **kw) -> ContestPage:
  """
    This method will use ``http`` to request ``/contest/<contest_id>``, and parse to struct result

    :param http: AioHttpHelperInterface 
    :param contest_id: contest id in url

    :returns: parsed structured result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.contest_meta import async_contest_meta

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          # you can login before request
          result = await async_contest_meta(http=http, contest_id = '1779')
          print(result.id)
          print(result.title)
          print(result.url)
          print(result.problems)
          print(result.materials)
          await http.close_session()

        asyncio.run(demo())
  """

  url = f"/contest/{contest_id}"
  resp = await http.async_get(url)
  soup = BeautifulSoup(resp, 'lxml')
  ogtitle = soup_find_bs4Tag(soup, 'meta', property="og:title")['content']
  assert isinstance(ogtitle, str)
  title = ogtitle[len('Dashboard -'):-len('- Codeforces')].strip()
  problems = parse_problems(resp)
  for problem in problems:
    problem.contest_id = contest_id

  return ContestPage(
      id=contest_id,
      # resp = resp,
      url=url,
      title=title,
      problems=problems,
      # questions=[],  # TODO
      materials=parse_materials(resp))



from dataclasses import dataclass
from enum import Enum
from typing import List, cast
from lxml import html
from lxml.etree import _Element

from .interfaces.AioHttpHelper import AioHttpHelperInterface
from . import util


class RegisterResultMsg(str, Enum):
  HaveBeenRegistered = 'You have been successfully registered'  # first time message
  AlreadyRegistered = 'You are already registered for the contest'
  NoRegistrationIsOpenedNow = 'No registration is opened now'
  Empty = ''


@dataclass
class RegisterResult:
  title: str = ''
  msg: RegisterResultMsg = RegisterResultMsg.Empty  # example: already register, TODO 有些比赛会 None


# TODO 统一参数命名
# Take part as individual participant
async def async_register(http: AioHttpHelperInterface, contest_id: str, **kw) -> RegisterResult:
  """
    This method will use ``http`` for register request, you need login before

    :param contest_id: Codeforces contest_id in url

    :returns: the result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.account import async_login
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.contest_register import async_register, RegisterResultMsg

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          result = await async_login(http=http, handle='<handle>', password='<password>')
          assert(result.success)
          result = await async_register(http=http,contest_id='1811')
          print(result)
          # assert result.msg == RegisterResultMsg.HaveBeenRegistered
          # assert result.msg == RegisterResultMsg.AlreadyRegistered
          # assert result.msg == RegisterResultMsg.NoRegistrationIsOpenedNow
          await http.close_session()

        asyncio.run(demo())
  """
  url = f'/contestRegistration/{contest_id}'
  resp = await http.async_get(url)
  doc = html.fromstring(resp)
  title = cast(List[_Element], doc.xpath('.//title'))[0].text
  msg = util.show_message(resp)  # AlreadyRegistered
  if msg:
    return RegisterResult(title=title, msg=msg)
  token = http.get_tokens()
  resp = await http.async_post(
      url,
      data=http.create_form({
          'csrf_token': token['csrf'],
          'action': 'formSubmitted',
          'takePartAs': 'personal',  # Take part as individual participant
          'backUrl': '',
      }))
  msg = util.show_message(resp)  # HaveBeenRegistered
  return RegisterResult(title=title, msg=msg)



import re
from typing import Tuple


def pid2url(problem_id: str) -> str:
  contest_id, problem_key = pid2split(problem_id)
  return f'/contest/{contest_id}/problem/{problem_key}'


def pid2split(problem_id) -> Tuple[str, str]:
  """
    problem id to [contest id, problem key]
    
    pid2split('1843F2') == ['1843','F2']
  """
  result = re.match('^(\\d+)([A-Z]\\d?)$', problem_id)
  if result is None:
    raise Exception('problem id[' + problem_id + '] ERROR')
  return str(result.group(1)), str(result.group(2))


def problem_url_parse(problem_url: str) -> Tuple[str, str]:
  """
    https://codeforces.com/contest/1740/problem/G

    convert to

    ['1740', 'G']
  """
  result = re.match('^.*contest/(.*)/problem/(.*)$', problem_url)
  assert result is not None
  return str(result.group(1)), str(result.group(2))


import logging


def has_level_handler(logger: logging.Logger) -> bool:
  """Check if there is a handler in the logging chain that will handle the
    given logger's :meth:`effective level <~logging.Logger.getEffectiveLevel>`.
    """
  level = logger.getEffectiveLevel()
  current = logger

  while current:
    if any(handler.level <= level for handler in current.handlers):
      return True

    if not current.propagate:
      break

    current = current.parent  # type: ignore

  return False


default_handler = logging.StreamHandler()  # type: ignore
default_handler.setFormatter(logging.Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s"))


# TODO cache
def create_logger(name, debug=False) -> logging.Logger:
  logger = logging.getLogger(name)

  if debug and not logger.level:
    logger.setLevel(logging.DEBUG)

  if not has_level_handler(logger):
    logger.addHandler(default_handler)

  return logger



from abc import abstractmethod
from typing import Any, Callable, Dict, Protocol, Tuple, AsyncIterator
import aiohttp


class AioHttpHelperInterface(Protocol):

  @abstractmethod
  def create_form(self, form_data: Dict[str, Any]) -> aiohttp.FormData:
    raise NotImplementedError

  @abstractmethod
  async def async_get(self, url: str, **kwargs) -> str:
    raise NotImplementedError

  @abstractmethod
  async def async_post(self, url: str, data: Any, **kwargs) -> str:
    raise NotImplementedError

  @abstractmethod
  def update_tokens(self, csrf: str, ftaa: str, bfaa: str, uc: str, usmc: str) -> None:
    raise NotImplementedError

  @abstractmethod
  async def open_session(self) -> aiohttp.ClientSession:
    raise NotImplementedError

  @abstractmethod
  def get_tokens(self) -> Any:
    raise NotImplementedError

  @abstractmethod
  async def close_session(self) -> None:
    raise NotImplementedError

  # TODO move call back out
  @abstractmethod
  async def websockets(self, url: str, callback: Callable[[Any], Tuple[bool, Any]]) -> AsyncIterator[Any]:
    raise NotImplementedError
    # mypy type hint https://github.com/python/mypy/issues/5070
    yield 0



from dataclasses import dataclass
import logging

from .constants import LIB_NAME
from .logging import create_logger


@dataclass
class ExtractResult:
  logger: logging.Logger


def extract_common_kwargs(**kwargs) -> ExtractResult:
  # TODO env log level
  logger = kwargs['logger'] if 'logger' in kwargs else create_logger(LIB_NAME)
  return ExtractResult(logger=logger)



from typing import List, cast
import bs4
from lxml import html
from lxml.etree import ElementBase
from lxml.html import HtmlElement


def pop_element(t):
  text = t.text
  t.getparent().remove(t)
  return text


def show_message(resp):
  doc = html.fromstring(resp)
  for lines in doc.xpath('.//script[@type="text/javascript" and not(@src)]'):
    for l in lines.text.splitlines():
      if l.find('Codeforces.showMessage("') != -1:
        return l.split('"')[1]


# for mypy type check
def soup_find_bs4Tag(soup: bs4.Tag, *args, **kwargs) -> bs4.Tag:
  result = soup.find(*args, **kwargs)
  assert isinstance(result, bs4.Tag)
  return result


def typedxpath(el: HtmlElement, s: str) -> List[HtmlElement]:
  return cast(List[HtmlElement], el.xpath(s))



from dataclasses import dataclass
import logging
from typing import Any, List, cast
from lxml import html

from .interfaces.AioHttpHelper import AioHttpHelperInterface
from .kwargs import extract_common_kwargs
from .util import pop_element


def extract_testcases(tags):
  ret = []
  for i in tags:
    pop_element(i.xpath('.//div[@class="title"]')[0])
    divs = i.xpath('.//div[@class]')
    if len(divs) == 0:
      ret.append("\n".join([t.strip() for t in i.itertext()]))
    else:
      l = ''
      prev = divs[0].get('class')
      lines = []
      for d in divs:
        if d.get('class') == prev:
          l += d.text + '\n'
        else:
          lines.append(l)
          prev = d.get('class')
          l = d.text + '\n'
      if l: lines.append(l.strip() + '\n')
      ret.append("\n".join(lines))
  return ret


@dataclass
class TestCase:
  in_data: str
  out_data: str


@dataclass
class ProblemInfo:
  # testcases: List[TestCase]
  title: str
  level: str
  time_limit_seconds: str
  memory_limit_mb: str
  desc: str
  in_tc: List[str]
  out_tc: List[str]
  note: str


async def async_problems(http: AioHttpHelperInterface, contest_id: str, **kw) -> List[ProblemInfo]:
  """
    This method will use ``http`` to request ``/contest/<contest_id>/problems``, and parse to struct result

    :param http: AioHttpHelperInterface 
    :param contest_id: contest id in url

    :returns: parsed structured result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.problems import async_problems

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          # you can login before request
          result = await async_problems(http=http, contest_id='1779')
          print(len(result))
          print(result[0])
          await http.close_session()

        asyncio.run(demo())
  """
  logger = extract_common_kwargs(**kw).logger
  url = "/contest/{}/problems".format(contest_id)
  resp = await http.async_get(url)
  doc = html.fromstring(resp)
  probs = cast(List[Any], doc.xpath('.//div[@class="problemindexholder"]'))

  ret: List[ProblemInfo] = []
  for p in probs:
    try:
      # if alert: alert = alert[0].text
      level = p.get('problemindex')
      typo = p.xpath('.//div[@class="ttypography"]')[0]
      title = pop_element(typo.xpath('.//div[@class="title"]')[0])
      time_limit = typo.xpath('.//div[@class="time-limit"]')[0]
      time_limit = [t for t in time_limit.itertext()][1].split(' ')[0]
      memory_limit = typo.xpath('.//div[@class="memory-limit"]')[0]
      memory_limit = [t for t in memory_limit.itertext()][1].split(' ')[0]
      desc = typo.xpath('.//div[not(@class)]')
      if desc:
        desc = '\n'.join([t for t in desc[0].itertext()])
      else:
        desc = ""

      for j in typo.xpath('.//div[@class="section-title"]'):
        pop_element(j)

      in_spec = typo.xpath('.//div[@class="input-specification"]')
      if in_spec:
        in_spec = '\n'.join([t for t in in_spec[0].itertext()])
      else:
        in_spec = ""

      out_spec = typo.xpath('.//div[@class="output-specification"]')
      if out_spec:
        out_spec = '\n'.join([t for t in out_spec[0].itertext()])
      else:
        out_spec = ""

      in_tc = extract_testcases(typo.xpath('.//div[@class="input"]'))
      out_tc = extract_testcases(typo.xpath('.//div[@class="output"]'))
      note = typo.xpath('.//div[@class="note"]')
      if note:
        note = '\n'.join([t for t in note[0].itertext()])
      ret.append(
          ProblemInfo(title=title,
                      level=level,
                      time_limit_seconds=time_limit,
                      memory_limit_mb=memory_limit,
                      desc=desc,
                      in_tc=in_tc,
                      out_tc=out_tc,
                      note=note))
    except Exception as e:
      logger.exception(e)
  return ret



from dataclasses import dataclass
from typing import Optional, Tuple
from random import choice
from lxml import html
from lxml.html import HtmlElement
import logging

from .kwargs import extract_common_kwargs
from .util import typedxpath
from .interfaces.AioHttpHelper import AioHttpHelperInterface

default_login_url = "/enter?back=%2F"


@dataclass
class LoginResult:
  html: str = ''
  csrf: str = ''
  ftaa: str = ''
  bfaa: str = ''
  uc: str = ''  # user channel ?
  usmc: str = ''
  # cc: str = ''  # contest channel? TODO remove, 这个和 contest相关, 不应该和login在一起
  # pc: str = ''  # remove same reason
  success: bool = False


def is_user_logged_in(html_data: str) -> bool:
  doc = html.fromstring(html_data)
  links = typedxpath(doc, './/div[@class="lang-chooser"]/div[not(@style)]/a[@href]')
  for m in links:
    if m.text.strip() in ["Register", "Enter"]:
      return False
  return True


async def async_fetch_logged_in(http: AioHttpHelperInterface, login_url=default_login_url, **kw) -> Tuple[bool, str]:
  """
    auto update token 
    return bool(is_logged_in), html_data
  """
  logger = extract_common_kwargs(**kw).logger

  html_data = await http.async_get(login_url)
  uc, usmc, cc, pc, csrf_token, ftaa, bfaa = extract_channel(html_data, logger=logger)

  if is_user_logged_in(html_data=html_data):
    http.update_tokens(csrf=csrf_token, ftaa=ftaa, bfaa=bfaa, uc=uc, usmc=usmc)
    return True, html_data
  return False, ''


# No exception, handler inside
def extract_channel(html_data: str,
                    logger: Optional[logging.Logger] = None) -> Tuple[str, str, str, str, str, str, str]:
  doc = html.fromstring(html_data)

  def xpath_content(el: HtmlElement, s: str) -> str:
    try:
      l = typedxpath(el, s)
      return l[0].get('content') if len(l) > 0 else ''
    except Exception as e:
      if logger: logger.exception(e)
      return ''

  uc = xpath_content(doc, './/meta[@name="uc"]')
  usmc = xpath_content(doc, './/meta[@name="usmc"]')
  cc = xpath_content(doc, './/meta[@name="cc"]')
  pc = xpath_content(doc, './/meta[@name="pc"]')
  try:
    csrf_token = typedxpath(doc, './/span[@class="csrf-token"]')[0].get('data-csrf')
    assert len(csrf_token) == 32, "Invalid CSRF token"
  except Exception as e:
    if logger: logger.exception(e)
    csrf_token = ''
  ftaa = ''.join([choice('abcdefghijklmnopqrstuvwxyz0123456789') for x in range(18)])
  # bfaa : Fingerprint2.x64hash128
  bfaa = ''.join([choice('0123456789abcdef') for x in range(32)])
  return uc, usmc, cc, pc, csrf_token, ftaa, bfaa


# TODO 已经登陆账号A, 再调用登陆账号B是不行的, 这个逻辑应该是由外部控制，调用时应该确保未登录状态
async def async_login(http: AioHttpHelperInterface,
                      handle: str,
                      password: str,
                      login_url=default_login_url,
                      **kw) -> LoginResult:
  """
    This method will use ``http`` for login request, and  :py:func:`is_user_logged_in()` for login check

    :param handle: Codeforces handle
    :param password: Codeforces password

    :returns: if it is successful post and logged

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.account import async_login, is_user_logged_in
        from codeforces_core.httphelper import HttpHelper

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          result = await async_login(http=http, handle='<handle>', password='<password>')
          assert(result.success)

          html_data = await http.async_get('https://codeforces.com')
          assert(is_user_logged_in(html_data))

          await http.close_session()

        asyncio.run(demo())
  """
  logger = extract_common_kwargs(**kw).logger
  html_data = await http.async_get(login_url)
  csrf_token, ftaa, bfaa = extract_channel(html_data, logger=logger)[4:7]
  login_data = {
      'csrf_token': csrf_token,
      'action': 'enter',
      'ftaa': ftaa,
      'bfaa': bfaa,
      'handleOrEmail': handle,
      'password': password,
      'remember': 'on',
  }
  html_data = await http.async_post(login_url, login_data)
  # uc, usmc, cc, pc, csrf_token, ftaa, bfaa = extract_channel(html_data)
  uc, usmc, cc, pc = extract_channel(html_data, logger=logger)[0:4]

  success = False
  # if check_login(result.html):
  if is_user_logged_in(html_data=html_data):
    http.update_tokens(csrf=csrf_token, ftaa=ftaa, bfaa=bfaa, uc=uc, usmc=usmc)
    success = True
  else:
    success = False

  return LoginResult(
      html=html_data,
      csrf=csrf_token,
      ftaa=ftaa,
      bfaa=bfaa,
      uc=uc,
      usmc=usmc,
      # cc=cc,
      # pc=pc,
      success=success)



from dataclasses import dataclass, field
from enum import Enum
from typing import List
from bs4 import BeautifulSoup
import bs4

from .interfaces.AioHttpHelper import AioHttpHelperInterface


@dataclass
class TestCase:
  in_data: str
  out_data: str


@dataclass
class ProblemInfo:
  # testcases: List[TestCase]
  title: str
  level: str
  time_limit_seconds: str
  memory_limit_mb: str
  desc: str
  in_tc: List[str]
  out_tc: List[str]
  note: str


@dataclass
class ParseProblemResult(object):

  class Status(str, Enum):
    AC = 'AC'
    FAILED = 'FAILED'
    NOTVIS = 'NOTVIS'

  status: Status = Status.NOTVIS
  title: str = ''
  test_cases: List[TestCase] = field(default_factory=lambda: [])
  id: str = ''
  oj: str = ''
  description: str = ''
  time_limit: str = ''
  mem_limit: str = ''
  url: str = ''
  html: str = ''
  file_path: str = ''


async def async_problem(http: AioHttpHelperInterface, contest_id: str, level: str, **kw) -> ParseProblemResult:
  """
    This method will use ``http`` to request ``/contest/<contest_id>/problems``, and parse to struct result

    :param http: AioHttpHelperInterface 
    :param contest_id: contest id in url

    :returns: parsed structured result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.problem import async_problem

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          # you can login before request
          result = await async_problem(http=http, contest_id='1779', level='F')
          print(result)
          await http.close_session()

        asyncio.run(demo())
  """
  resp = await http.async_get(f'/contest/{contest_id}/problem/{level}')
  problem = ParseProblemResult(html=resp)
  soup = BeautifulSoup(resp, 'lxml')

  # TODO implememt soup_find function to assert type for mypy
  match_groups = soup.find('div', attrs={'class': 'title'})
  assert isinstance(match_groups, bs4.Tag)
  problem.title = str(match_groups.string)[2:].strip(" \r\n")

  match_groups = soup.find(name='div', attrs={'class': 'time-limit'})
  assert isinstance(match_groups, bs4.Tag)
  problem.time_limit = str(match_groups.contents[-1]).strip()

  match_groups = soup.find(name='div', attrs={'class': 'memory-limit'})
  assert isinstance(match_groups, bs4.Tag)
  problem.mem_limit = str(match_groups.contents[-1]).strip()

  match_groups = soup.find(name='div', attrs={'class': 'problem-statement'})

  problem.status = ParseProblemResult.Status.NOTVIS  # TODO for show progress

  match_groups = soup.find(name='div', attrs={'class': 'sample-test'})
  assert isinstance(match_groups, bs4.Tag)
  problem.test_cases.clear()
  if match_groups:
    test_case_inputs = match_groups.find_all(name='div', attrs={'class': 'input'})
    test_case_outputs = match_groups.find_all(name='div', attrs={'class': 'output'})
    assert (len(test_case_inputs) == len(test_case_outputs))  # may not? in April fool contest
    for i in range(len(test_case_inputs)):
      t_in = test_case_inputs[i].find(name='pre').get_text("\n").strip(" \r\n")
      t_out = test_case_outputs[i].find(name='pre').get_text("\n").strip(" \r\n")
      problem.test_cases.append(TestCase(t_in, t_out))

  return problem



import logging
from typing import Any, AsyncIterator, Callable, Dict, List, Optional, Tuple, cast
from lxml import html
from lxml.etree import _Element
from os import path
import asyncio
import aiohttp
import pyaes
import json
import re

from .constants import CF_HOST
from .interfaces.AioHttpHelper import AioHttpHelperInterface
from .kwargs import extract_common_kwargs

default_headers = {
    'Accept': '*/*',
    'Accept-Encoding': 'gzip',
    # 'User-Agent': config.conf['user_agent'], TODO
    'User-Agent':
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36'
}


class RCPCRedirectionError(Exception):

  def __init__(self):
    super().__init__("RCPC redirection detected")


def add_header(newhdr, headers: Dict[str, str]) -> Dict[str, str]:
  headers.update(newhdr)
  return headers


async def on_request_end(session, trace_request_ctx, params):
  elapsed = asyncio.get_event_loop().time() - trace_request_ctx.start
  print("[*] Request end : {}".format(elapsed))


class HttpHelper(AioHttpHelperInterface):
  session: Optional[aiohttp.ClientSession] = None
  cookie_jar_path = ''
  cookie_jar: Optional[aiohttp.CookieJar] = None
  token_path = ''
  tokens: Dict[str, str] = {}
  headers: Dict[str, str] = {}  # TODO
  logger: logging.Logger

  def __init__(self,
               cookie_jar_path: str = '',
               token_path: str = '',
               headers=default_headers,
               host=CF_HOST,
               **kw) -> None:
    # if path is empty string then won't save to any file, just store in memory
    self.cookie_jar_path = cookie_jar_path
    # if path is empty string then won't save to any file, just store in memory
    self.token_path = token_path
    self.headers = headers
    # TODO support cf mirror site?
    self.host = host
    self.logger = extract_common_kwargs(**kw).logger

  @staticmethod
  def load_tokens(token_path: str) -> Dict[str, Any]:
    if token_path and path.isfile(token_path):
      with open(token_path, 'r') as f:
        return json.load(f)
    return {}

  @staticmethod
  def load_cookie_jar(cookie_jar_path: str) -> aiohttp.CookieJar:
    jar = aiohttp.CookieJar()
    if cookie_jar_path:
      if path.isfile(cookie_jar_path):
        jar.load(file_path=cookie_jar_path)
      else:
        jar.save(file_path=cookie_jar_path)
    return jar

  async def open_session(self) -> aiohttp.ClientSession:
    self.cookie_jar = HttpHelper.load_cookie_jar(self.cookie_jar_path)
    self.tokens = HttpHelper.load_tokens(self.token_path)
    self.session = await aiohttp.ClientSession(cookie_jar=self.cookie_jar).__aenter__()
    return self.session

  async def close_session(self) -> None:
    await self.session.__aexit__(None, None, None)
    self.tokens = {}
    self.cookie_jar = None
    self.session = None

  def update_tokens(self, csrf: str, ftaa: str, bfaa: str, uc: str, usmc: str) -> None:
    self.tokens = {'csrf': csrf[:32], 'ftaa': ftaa, 'bfaa': bfaa, 'uc': uc, 'usmc': usmc}
    if self.token_path:
      with open(self.token_path, 'w') as f:
        json.dump(self.tokens, f)

  async def async_get(self, url, headers=None, csrf=False):
    if self.session is None:
      raise Exception('Please open_session() before async_get()')
    if headers == None: headers = default_headers
    if csrf and 'csrf' in self.tokens:
      headers = add_header({'X-Csrf-Token': self.tokens['csrf']}, headers=headers)
    # TODO remove the feature
    if url.startswith('/'): url = self.host + url
    try:
      async with self.session.get(url, headers=headers) as response:
        assert response.status == 200
        text = await response.text()
        self.check_rcpc(text)
        if self.cookie_jar_path:
          self.cookie_jar.save(file_path=self.cookie_jar_path)  # TODO move auto save to file out
        return text
    except RCPCRedirectionError:
      async with self.session.get(url, headers=headers) as response:
        assert response.status == 200
        if self.cookie_jar_path:
          self.cookie_jar.save(file_path=self.cookie_jar_path)
        return await response.text()
    except Exception as e:
      self.logger.error(e)

  async def async_post(self, url, data, headers=default_headers, csrf=False, **kwargs: Any):
    if self.session is None:
      raise Exception('Please open_session() before async_get()')
    if headers == None: headers = default_headers
    if csrf and 'csrf' in self.tokens:
      headers = add_header({'X-Csrf-Token': self.tokens['csrf']}, headers=headers)

    # TODO remove the feature
    if url.startswith('/'): url = self.host + url
    try:
      async with self.session.post(url, headers=headers, data=data, **kwargs) as response:
        assert response.status == 200
        self.check_rcpc(await response.text())
        if self.cookie_jar_path:
          self.cookie_jar.save(file_path=self.cookie_jar_path)
        return await response.text()
    except RCPCRedirectionError:
      async with self.session.post(url, headers=headers, data=data) as response:
        assert response.status == 200
        if self.cookie_jar_path:
          self.cookie_jar.save(file_path=self.cookie_jar_path)
        return await response.text()
    except Exception as e:
      self.logger.error(e)

  def get_tokens(self):
    return self.tokens

  def check_rcpc(self, html_data: str):
    doc = html.fromstring(html_data)
    aesmin = cast(List[_Element], doc.xpath(".//script[@type='text/javascript' and @src='/aes.min.js']"))
    if len(aesmin) > 0:
      print("[+] RCPC redirection detected")
      js = cast(List[_Element], doc.xpath(".//script[not(@type)]"))
      assert len(js) > 0
      keys = re.findall(r'[abc]=toNumbers\([^\)]*', js[0].text)
      for k in keys:
        if k[0] == 'a':
          key = bytes.fromhex(k.split('"')[1])
        elif k[0] == 'b':
          iv = bytes.fromhex(k.split('"')[1])
        elif k[0] == 'c':
          ciphertext = bytes.fromhex(k.split('"')[1])
      assert len(key) == 16 and len(iv) == 16 and len(ciphertext) == 16, 'AES decryption error'
      c = pyaes.AESModeOfOperationCBC(key, iv=iv)
      plaintext = c.decrypt(ciphertext)
      rcpc = plaintext.hex()
      self.cookie_jar.update_cookies({'RCPC': rcpc})
      self.cookie_jar.save(file_path=self.cookie_jar_path)
      raise RCPCRedirectionError()

  def create_form(self, form_data) -> aiohttp.FormData:
    form = aiohttp.FormData()
    for k, v in form_data.items():
      form.add_field(k, v)
    return form

  # callback return (end watch?, transform result)
  async def websockets(self, url: str, callback: Callable[[Any], Tuple[bool, Any]]) -> AsyncIterator[Any]:
    try:
      async with self.session.ws_connect(url) as ws:
        async for msg in ws:
          if msg.type == aiohttp.WSMsgType.TEXT:
            js = json.loads(msg.data)
            js['text'] = json.loads(js['text'])

            endwatch, obj = callback(js)
            yield obj
            if endwatch:
              return

          else:
            self.logger.error('wrong msg type?', msg.type)
            break
        return
    except Exception as e:
      self.logger.error(e)
      return



from time import time
from typing import Any, Callable, Tuple, AsyncIterator

from .account import extract_channel
from .interfaces.AioHttpHelper import AioHttpHelperInterface
from .kwargs import extract_common_kwargs
from .submit import SubmissionWSResult


# return (end watch?, transform result)
def display_ws(result: Any) -> Tuple[bool, Any]:
  print(result)
  return False, result


# # {'id': 1, 'channel': '34f1ec4b729022e4b48f8d24b65c857805a90469', 'text': {'t': 's', 'd': [5973356517882654806, 200631363, 1777, 1746206, 'TESTS', None, 'OK', 86, 86, 3198, 7884800, 148217099, '21220', '04.04.2023 5:57:08', '04.04.2023 5:57:08', 2147483647, 73, 0]}}
# # 总的ws, 无法获得当前题目的 通过百分比
# def create_ws_task(http: AioHttpHelperInterface, ws_handler: Callable[[Any], Tuple[bool, Any]]) -> asyncio.Task:
#   """
#     This method will use ``http`` to create common websocket, and ``ws_handler`` to handle each ws message
#
#     this websocket cannot receive a submission running percentage, use :py:func:`create_contest_ws_task()` instead
#
#     :param http: AioHttpHelperInterface
#     :param ws_handler: function to handler messages
#
#     :returns: the task which run ws
#
#     Examples:
#
#     .. code-block::
#
#         import asyncio
#         from codeforces_core.httphelper import HttpHelper
#         from codeforces_core.account import async_login
#         from codeforces_core.websocket import create_ws_task, display_ws
#         from codeforces_core.submit import async_submit
#
#         async def demo():
#           # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
#           http = HttpHelper(token_path='', cookie_jar_path='')
#           await http.open_session()
#           result = await async_login(http=http, handle='<handle>', password='<password>')
#           assert(result.success)
#           task = create_ws_task(http, ws_handler=display_ws)
#           # submit code in webpage
#           try:
#             result = await asyncio.wait_for(task, timeout=60)
#             print("ws is done, result:", task.result())
#           except asyncio.TimeoutError:
#             pass
#           await http.close_session()
#
#         asyncio.run(demo())
#   """
#   epoch = int(time() * 1000)  # s -> ms
#   token = http.get_tokens()
#   ws_url = f"wss://pubsub.codeforces.com/ws/{token['uc']}/{token['usmc']}?_={epoch}&tag=&time=&eventid="
#   print(ws_url)
#   return asyncio.create_task(http.websockets(ws_url, ws_handler))


# https://codeforces.com/contest/<contest_id>/my 会多出两个 meta
#    <meta name="cc" content="xxx"/>
#    <meta name="pc" content="yyy"/>
#   TODO 设计上不太对, handler处理了数据, 结果也抛给了使用者, 应该handler 只关心是否停止, 而transform不应该在handler里处理
# 这两个可以监听 题目测试时 的通过 百分比 变化
async def create_contest_ws_task_yield(http: AioHttpHelperInterface, contest_id: str,
                                       ws_handler: Callable[[Any], Tuple[bool, Any]],
                                       **kw) -> AsyncIterator[SubmissionWSResult]:
  """
    This method will use ``http`` to create contest specific websocket, and ``ws_handler`` to handle each ws message

    :param http: AioHttpHelperInterface
    :param contest_id: contest id in the url
    :param ws_handler: function to handler messages

    :returns: the task which run ws

    Examples:

    See docstring of :py:func:`codeforces_core.submit.async_submit()`
  """
  logger = extract_common_kwargs(**kw).logger
  epoch = int(time() * 1000)  # s -> ms
  html_data = await http.async_get(f"/contest/{contest_id}/my")
  cc, pc = extract_channel(html_data, logger)[2:4]
  assert cc and pc
  ws_url = f"wss://pubsub.codeforces.com/ws/s_{pc}/s_{cc}?_={epoch}&tag=&time=&eventid="
  logger.debug(f"pc = {pc}")  # 似乎和场次有关, 可能包含别人的?
  logger.debug(f"cc = {cc}")  # 似乎只会包含自己的
  logger.debug(f"ws_url = {ws_url}")
  async for data in http.websockets(ws_url, ws_handler):
    yield data



__all__ = (
    "interfaces",
    "account",
    "contest_list",
    "contest_meta",
    "contest_register",
    "contest_standing",
    "httphelper",
    "language",
    "problem",
    "problems",
    "submit",
    "util",
    "websocket",
)
__version__ = "0.0.2.2"



from dataclasses import dataclass, field
import re
from typing import List, Optional
from bs4 import BeautifulSoup
import bs4

from .interfaces.AioHttpHelper import AioHttpHelperInterface

# TODO post toggle showunofficial


@dataclass
class StandingProblem:
  id: str = ''
  score: str = ''
  time: str = ''


@dataclass
class StandingRow:
  rank: str = ''
  who: str = ''
  passed: Optional[str] = None
  score: str = ''
  hack: str = ''
  penalty: str = ''
  problems: List[StandingProblem] = field(default_factory=lambda: [])


@dataclass
class Standing:
  head: List[str]
  rows: List[StandingRow]
  url: str = ''


def parseStandingHtml(html) -> Standing:
  soup = BeautifulSoup(html, 'lxml')
  currentContestList = soup.find('div', class_='datatable')
  assert isinstance(currentContestList, bs4.Tag)
  rows: List[StandingRow] = []
  trs: List[BeautifulSoup] = currentContestList.find_all('tr')
  ths: List[BeautifulSoup] = trs[0].find_all('th')
  h: List[str] = ['' for i in ths]  # head: get info from th
  for i in range(len(ths)):
    text = ths[i].get_text().strip()
    if text == '#':
      h[i] = 'rank'
    elif text == 'Who':
      h[i] = 'who'
    elif text == '=':
      h[i] = 'score'
    elif text == '*':
      h[i] = 'hack'
    elif text == 'Penalty':
      h[i] = 'penalty'
    else:
      a = ths[i].find('a')
      if a is not None:
        h[i] = a.get_text().strip()

  for i in range(1, len(trs) - 1):  # ignore first(head) and last line(total accepted TODO)
    tds = trs[i].find_all('td')
    row = StandingRow()
    for j in range(len(h)):
      if h[j] == 'rank':
        row.rank = tds[j].get_text().strip()
        # '4\xa0(13)' => '13'
        inbracket = re.search("\\(([0-9]*)\\)", row.rank)
        if inbracket:
          row.rank = inbracket.group(1)
      elif h[j] == 'who':
        row.who = tds[j].get_text().strip()
      elif h[j] == 'score':
        row.score = tds[j].get_text().strip()
      elif h[j] == 'hack':
        row.hack = tds[j].get_text().strip()
      elif h[j] == 'penalty':
        row.penalty = tds[j].get_text().strip()
      else:  # problems
        passScore = tds[j].find_all('span', class_='cell-passed-system-test')
        problem = StandingProblem(id=h[j])
        if passScore and len(passScore) > 0:
          problem.score = passScore[0].get_text().strip()
          problem.time = tds[j].find('span', class_="cell-time").get_text().strip()
        else:
          problem.score = tds[j].get_text().strip()
        row.problems.append(problem)
    rows.append(row)

  # TODO add pagenation
  return Standing(head=h, rows=rows)


async def async_friends_standing(http: AioHttpHelperInterface, contest_id: str, **kw) -> Standing:
  """
    This method will use ``http`` to request ``/contest/<contest_id>/standings/friends/true``, and parse to struct result

    :param contest_id: contest id in url

    :returns: parsed structured result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.account import async_login
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.contest_standing import async_friends_standing

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          result = await async_login(http=http, handle='<handle>', password='<password>')
          assert(result.success)
          result = await async_friends_standing(http=http, contest_id='1779')
          print(result.head)
          print(len(result.rows))
          print(result.rows[3])
          await http.close_session()

        asyncio.run(demo())
  """
  url = f'/contest/{contest_id}/standings/friends/true'
  resp = await http.async_get(url)
  result = parseStandingHtml(resp)
  result.url = url
  return result


async def async_common_standing(http: AioHttpHelperInterface, contest_id: str, page: str = '1', **kw) -> Standing:
  """
    This method will use ``http`` to request ``/contest/<contest_id>/standings/page/<page>``, and parse to struct result

    :param contest_id: contest id in url
    :param page: pagination 1-index

    :returns: parsed structured result 

    Examples:

    .. code-block::

        import asyncio
        from codeforces_core.httphelper import HttpHelper
        from codeforces_core.contest_standing import async_common_standing

        async def demo():
          # http = HttpHelper(token_path='/tmp/cache_token', cookie_jar_path='/tmp/cache_cookie_jar')
          http = HttpHelper(token_path='', cookie_jar_path='')
          await http.open_session()
          result = await async_common_standing(http=http, contest_id='1779', page='2')
          print(result.head)
          print(len(result.rows))
          print(result.rows[3])
          await http.close_session()

        asyncio.run(demo())
  """
  url = f'/contest/{contest_id}/standings/page/{page}'
  resp = await http.async_get(url)
  result = parseStandingHtml(resp)
  result.url = url
  return result
